from .execution_logs import execution_logs_repository_factory
from .notifications import notification_repository_factory
from .connectors import connectors_api_http_client_factory
from .requirements import requirements_repository_factory
from .stage_run import stage_run_repository_factory
from .execution import execution_repository_factory
from .tables import tables_api_http_client_factory
from .env_vars import env_vars_repository_factory
from .authn import authn_repository_factory
from .lock import lock_repository_factory


connectors_api_http_client = connectors_api_http_client_factory()
execution_logs_repository = execution_logs_repository_factory()
requirements_repository = requirements_repository_factory()
notification_repository = notification_repository_factory()
tables_api_http_client = tables_api_http_client_factory()
stage_run_repository = stage_run_repository_factory()
execution_repository = execution_repository_factory()
env_vars_repository = env_vars_repository_factory()
authn_repository = authn_repository_factory()
lock_repository = lock_repository_factory()


__all__ = [
    "connectors_api_http_client",
    "execution_logs_repository",
    "notification_repository",
    "tables_api_http_client",
    "stage_run_repository",
    "execution_repository",
    "env_vars_repository",
    "authn_repository",
    "lock_repository",
]



from ....utils import serialize as json_serialize
import typing
from datetime import datetime, date
from uuid import UUID


def escape_ref(ref: str):
    return ref.replace('"', '""')


def quoted_identifier(ref: str):
    return f'"{escape_ref(ref)}"'


def serialize(value: typing.Any):
    if isinstance(value, (datetime, date)):
        return value.isoformat()
    if isinstance(value, (dict, list, tuple)):
        return json_serialize(value)
    if isinstance(value, set):
        return json_serialize(list(value))
    if isinstance(value, UUID):
        return str(value)

    return value



import io, tempfile, shutil, uuid, typing as t, pathlib
from werkzeug.datastructures import FileStorage


def public_path(name: str):
    return "/_files/" + name


def ensure_uploaded_files_dir():
    uploaded_files_dir = pathlib.Path(tempfile.gettempdir(), "_uploaded_files")
    uploaded_files_dir.mkdir(exist_ok=True)
    return uploaded_files_dir


def internal_path(name: str):
    uploaded_files_dir = ensure_uploaded_files_dir()
    return uploaded_files_dir.joinpath(name)


def get_random_filepath(name=None):
    if name is None:
        name = str(uuid.uuid4())
    else:
        name = str(uuid.uuid4()) + "/" + pathlib.Path(name).name
    path = internal_path(name)
    path.parent.mkdir(exist_ok=True)
    return name, path


def upload_file(file: t.Union[FileStorage, io.IOBase, pathlib.Path]):
    if isinstance(file, (io.BufferedReader, io.TextIOWrapper)):
        # TODO: use hasattr(file, "name")
        name, path = get_random_filepath(file.name)
        shutil.copy(file.name, path)
        return public_path(name)

    if isinstance(file, FileStorage):
        name, path = get_random_filepath(file.name)
        file.save(path)
        return public_path(name)

    if isinstance(file, pathlib.Path):
        name, path = get_random_filepath(file.name)
        shutil.copy(file, path)
        return public_path(name)

    if isinstance(file, io.IOBase):
        name, path = get_random_filepath()
        with open(path, "wb") as f:
            shutil.copyfileobj(file, f)
        return public_path(name)

    raise ValueError(f"Cannot upload {type(file)}")



from ..utils import serialize
from typing import TypeVar, Dict

T = TypeVar("T")


class SerializationHelper:
    KB = 1024
    MAX_SIZE = 100 * KB

    @classmethod
    def enforce_max_size(cls, obj: Dict) -> Dict[str, str]:
        try:
            dump = serialize(obj)
            if len(dump) > cls.MAX_SIZE:
                return {"serialization-error": "payload too large"}
            return obj
        except Exception:
            return {"serialization-error": "serialization failed"}



from unittest import TestCase

from ...repositories.project.project import ConditionStage, WorkflowTransition
from ...repositories.stage_run import LocalStageRunRepository
from . import workflow_engine


class ConditionTransitionsTest(TestCase):
    def setUp(self) -> None:
        self.repository = LocalStageRunRepository()

    def test_not_match_when_no_transitions(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        stage_run = self.repository.create_initial(stage="1", data={"foo": "bar"})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [])

    def test_not_match_when_variable_not_found(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="foo",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [])

    def test_match_when_equal(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="bar",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": "bar"})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": "bar"})])

    def test_nested_value(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo.bar",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="afulepa",
                )
            ],
        )
        stage_run = self.repository.create_initial(
            stage="1", data={"foo": {"bar": "afulepa"}}
        )
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(
            result, [dict(stage="target_id", data={"foo": {"bar": "afulepa"}})]
        )

    def test_match_str_when_int(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="1",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": 1})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": 1})])

    def test_match_none_when_not_found(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value=None,
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={})])

    def test_match_none_when_none(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value=None,
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": None})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": None})])

    def test_match_empty_when_none(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": None})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": None})])

    def test_match_empty_when_not_found(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={})])

    def test_match_empty_when_empty(self):
        stage = ConditionStage(
            id="1",
            title="title",
            variable_name="foo",
            workflow_position=(0, 0),
            workflow_transitions=[
                WorkflowTransition(
                    id="1",
                    target_type="target_type",
                    target_id="target_id",
                    type="conditions:patternMatched",
                    condition_value="",
                )
            ],
        )
        stage_run = self.repository.create_initial(stage="1", data={"foo": ""})
        result = workflow_engine._follow_condition_transitions(stage, stage_run)
        self.assertEqual(result, [dict(stage="target_id", data={"foo": ""})])



from unittest import TestCase
from abstra_internals.utils import serialize
from tests.fixtures import init_dir, clear_dir
from .launch_json import configure_launch_json, is_launch_json_configured


class ConfigureLaunchJsonTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_add_json_when_empty(self):
        self.assertFalse(is_launch_json_configured())

        configure_launch_json()

        self.assertTrue(is_launch_json_configured())

    def test_add_json_when_not_empty(self):
        launch_json_path = self.root.joinpath(".vscode/launch.json")
        launch_json_path.parent.mkdir(exist_ok=True, parents=True)
        launch_json_path.write_text(
            serialize({"version": "0.2.0", "configurations": []}), encoding="utf-8"
        )

        self.assertFalse(is_launch_json_configured())

        configure_launch_json()

        self.assertTrue(is_launch_json_configured())

    def test_do_nothing_when_already_configured(self):
        configure_launch_json()

        launch_json_path = self.root.joinpath(".vscode/launch.json")
        last_modified = launch_json_path.stat().st_mtime

        self.assertTrue(is_launch_json_configured())

        configure_launch_json()

        self.assertTrue(is_launch_json_configured())
        self.assertEqual(last_modified, launch_json_path.stat().st_mtime)


class IsLaunchJsonConfiguredTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_when_not_exists(self):
        self.assertFalse(is_launch_json_configured())

    def test_when_exists_but_no_config(self):
        launch_json_path = self.root.joinpath(".vscode/launch.json")
        launch_json_path.parent.mkdir(exist_ok=True, parents=True)
        launch_json_path.write_text(
            serialize({"version": "0.2.0", "configurations": []}), encoding="utf-8"
        )

        self.assertFalse(is_launch_json_configured())

    def test_when_exists_correctly(self):
        launch_json_path = self.root.joinpath(".vscode/launch.json")
        launch_json_path.parent.mkdir(exist_ok=True, parents=True)
        launch_json_path.write_text(
            serialize(
                {
                    "version": "0.2.0",
                    "configurations": [
                        {
                            "name": "Abstra debugger",
                            "type": "python",
                            "request": "attach",
                            "connect": {
                                "host": "localhost",
                                "port": "${input:abstraDebugPort}",
                            },
                        }
                    ],
                    "inputs": [{"id": "abstraDebugPort", "type": "promptString"}],
                }
            ),
            encoding="utf-8",
        )

        self.assertTrue(is_launch_json_configured())



import unittest
import ast
from typing import List
from .code import function_called_args, subscript_called_args

# Include the implementation of FunctionCallVisitor and function_called_args here (not repeated to save space)


class TestFunctionCalledArgs(unittest.TestCase):
    def test_import_from_direct_call(self):
        code = """
from os import getenv

getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_import_from_as_no_call(self):
        code = """
from os import getenv as alias

getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is None

    def test_import_from_as_call(self):
        code = """
from os import getenv as alias

alias("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_import_direct_call(self):
        code = """
import os

os.getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_import_as_no_call(self):
        code = """
import os as alias

os.getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is None

    def test_import_as_call(self):
        code = """
import os as alias

alias.getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_multiple_calls(self):
        code = """
import os as alias

alias.getenv("FOO")
alias.getenv("BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is not None
        assert len(result) == 2
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO"
        assert isinstance(result[1], list)
        assert len(result[1]) == 1
        assert isinstance(result[1][0], ast.Constant)
        assert result[1][0].value == "BAR"

    def test_empty_code(self):
        code = ""
        result = function_called_args(code, ["os"], "getenv")
        assert result is None

    def test_import_from_no_call(self):
        code = """
from os import getenv

os.getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is None

    def test_import_wildcard(self):
        code = """
from os import *

getenv("FOO_BAR")
"""
        result = function_called_args(code, ["os"], "getenv")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_nested_import_with_alias(self):
        code = """
import abstra.workflows as aw

aw.get_data("FOO_BAR")
"""
        result = function_called_args(code, ["abstra", "workflows"], "get_data")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"


class TestSubscriptCalledArgs(unittest.TestCase):
    def test_import_direct_subscript(self):
        code = """
import os

os.environ["FOO_BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_import_from_direct_subscript(self):
        code = """
from os import environ

environ["FOO_BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_import_as_subscript_no_call(self):
        code = """
import os as alias

os.environ["FOO_BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        # Since os is aliased to alias but os.environ is directly accessed
        # Depending on implementation, this might not be caught
        assert result is None

    def test_import_as_subscript_call(self):
        code = """
import os as alias

alias.environ["FOO_BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"

    def test_empty_code(self):
        code = ""
        result = subscript_called_args(code, ["os"], "environ")
        assert result is None

    def test_multiple_subscript_calls(self):
        code = """
import os as alias

alias.environ["FOO"]
alias.environ["BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        assert result is not None
        assert len(result) == 2
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO"
        assert isinstance(result[1], list)
        assert len(result[1]) == 1
        assert isinstance(result[1][0], ast.Constant)
        assert result[1][0].value == "BAR"

    def test_subscript_no_call(self):
        code = """
from os import environ

os.environ["FOO_BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        # If the setup does not interpret os.environ as a call to an imported name,
        # it won't return arguments for it
        assert result is None

    # This test is arguably not directly applicable as wildcards don't affect the way
    # subscript operations are identified in Python code the same way they do with function calls.
    # But if your SubscriptVisitor does handle wildcards in such a way:
    def test_import_wildcard(self):
        code = """
from os import *

environ["FOO_BAR"]
"""
        result = subscript_called_args(code, ["os"], "environ")
        assert result is not None
        assert len(result) == 1
        assert isinstance(result[0], list)
        assert len(result[0]) == 1
        assert isinstance(result[0][0], ast.Constant)
        assert result[0][0].value == "FOO_BAR"



from unittest import TestCase
from .unset_match_thread import UnsetMatchThread
from tests.fixtures import init_dir, clear_dir
from ...repositories.project.project import (
    ProjectRepository,
    ScriptStage,
    HookStage,
    WorkflowTransition,
)


class UnsetMatchThreadTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_unset_match_thread_valid_default(self):
        rule = UnsetMatchThread()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_valid_unset_match_thread(
        self,
    ):
        rule = UnsetMatchThread()
        project = ProjectRepository.load()

        hook_file = self.root / "hook.py"
        hook_file.write_text(
            "\n".join(
                [
                    "from abstra.workflows import match_thread",
                    "try:",
                    "    match_thread({'test': 0})",
                    "except:",
                    "    pass",
                ]
            )
        )
        hook = HookStage(
            id="test",
            file=str(hook_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
            enabled=True,
            path="hook",
        )
        project.add_stage(hook)

        script_file = self.root / "script.py"
        script_file.write_text(
            "\n".join(
                [
                    "import abstra.workflows as aw",
                    "aw.set_data('test', 1)",
                ]
            )
        )
        script = ScriptStage(
            id="test",
            file=str(script_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script.workflow_transitions = [
            WorkflowTransition(
                id="1",
                target_type=hook.type_name,
                target_id=hook.id,
                type=WorkflowTransition.default_type(script.type_name),
                condition_value=None,
            )
        ]

        project.add_stage(script)

        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 0)

    def test_invalid_unset_match_thread(
        self,
    ):
        rule = UnsetMatchThread()
        project = ProjectRepository.load()

        hook_file = self.root / "hook.py"
        hook_file.write_text(
            "\n".join(
                [
                    "from abstra.workflows import match_thread",
                    "try:",
                    "    match_thread({'unset': 0})",
                    "except:",
                    "    pass",
                ]
            )
        )
        hook = HookStage(
            id="test",
            file=str(hook_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
            enabled=True,
            path="hook",
        )
        project.add_stage(hook)

        script_file = self.root / "script.py"
        script_file.write_text(
            "\n".join(
                [
                    "import abstra.workflows as aw",
                    "aw.set_data('set', 1)",
                ]
            )
        )
        script = ScriptStage(
            id="test",
            file=str(script_file),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        script.workflow_transitions = [
            WorkflowTransition(
                id="1",
                target_type=hook.type_name,
                target_id=hook.id,
                type=WorkflowTransition.default_type(script.type_name),
                condition_value=None,
            )
        ]

        project.add_stage(script)

        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 1)



import json, os, shutil, sys, tempfile, uuid
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union
from pydantic.dataclasses import dataclass
from pathlib import Path

from . import json_migrations
from ...settings import Settings
from ...utils.file import traverse_code
from ...utils.string import to_kebab_case
from ...utils.format import normalize_path
from ...utils.graph import Edge, Graph, Node
from ...utils import check_is_url, nested_get


ServedStage = Union["FormStage", "HookStage"]
ActionStage = Union["FormStage", "HookStage", "JobStage", "ScriptStage"]
ControlStage = Union["IteratorStage", "ConditionStage"]
WorkflowStage = Union[ActionStage, ControlStage]


@dataclass
class NotificationTrigger:
    variable_name: str
    enabled: bool

    def validate_email(self, email: str) -> bool:
        return type(email) == str and "@" in email

    def get_recipients(self, thread_data: Dict[str, Any]) -> List[str]:
        variable_name = self.variable_name
        if not variable_name:
            return []

        raw_value = nested_get(thread_data, variable_name)
        if not raw_value:
            return []

        emails: List[str] = []
        if isinstance(raw_value, str):
            emails.append(raw_value)

        if isinstance(raw_value, list):
            emails = raw_value

        return [email for email in emails if self.validate_email(email)]

    def to_dto(self) -> Optional[Dict[str, Any]]:
        return {"variable_name": self.variable_name, "enabled": self.enabled}


@dataclass
class WorkflowTransition:
    id: str
    target_id: str
    target_type: str
    type: str
    condition_value: Optional[str] = None

    def __post_init__(self):
        self.id = str(uuid.uuid4()) if self.id is None else self.id

    def matches(self, variable_value):
        if self.type != "conditions:patternMatched":
            return False
        if type(variable_value) == int or type(variable_value) == float:
            return self.condition_value == str(variable_value)
        if type(variable_value) == str:
            return self.condition_value == variable_value
        elif variable_value is None:
            return self.condition_value is None or self.condition_value == ""
        else:
            print(f"Unknown type {type(variable_value)} for variable_value")
        return False

    @property
    def as_dict(self) -> dict:
        return {
            "target_id": self.target_id,
            "target_type": self.target_type,
            "type": self.type,
            "id": self.id,
            "condition_value": self.condition_value,
        }

    @staticmethod
    def from_dict(data: dict):
        if "type" not in data:
            raise Exception("Transition type is required")
        return WorkflowTransition(
            target_id=data["target_id"],
            target_type=data["target_type"],
            type=data["type"],
            id=data["id"],
            condition_value=data.get("condition_value"),
        )

    @staticmethod
    def default_type(source_type: Literal["form", "hook", "job", "script"]):
        if source_type == "form":
            return "forms:finished"
        elif source_type == "hook":
            return "hooks:finished"
        elif source_type == "job":
            return "jobs:finished"
        elif source_type == "script":
            return "scripts:finished"
        else:
            raise Exception(f"Invalid source type {source_type}")


@dataclass
class HookStage:
    id: str
    file: str
    path: str
    title: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    is_initial: bool = True
    enabled: bool = False

    @staticmethod
    def from_dict(dto: dict):
        x, y = dto["workflow_position"]
        return HookStage(
            id=dto["id"],
            file=dto["file"],
            path=dto["path"],
            title=dto["title"],
            enabled=dto["enabled"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in dto["transitions"]
            ],
            is_initial=dto["is_initial"],
        )

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id: str = id or str(uuid.uuid4())
        return HookStage(
            id=_id,
            file=file,
            path=to_kebab_case(file[:-3]),
            title=title,
            is_initial=True,
            workflow_transitions=[],
            workflow_position=workflow_position,
        )

    @property
    def type_name(self):
        return "hook"

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "path": self.path,
            "type": "hook",
            "is_initial": self.is_initial,
        }

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "file": self.file,
            "path": self.path,
            "title": self.title,
            "enabled": self.enabled,
            "workflow_position": self.workflow_position,
            "is_initial": self.is_initial,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["title", "enabled"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            elif attr == "path":
                setattr(self, attr, normalize_path(value))
            else:
                raise Exception(f"Cannot update {attr} of hook")

    def duplicate(self, new_id: str, new_position: Tuple[int, int] = (0, 0)):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )


@dataclass
class ScriptStage:
    id: str
    file: str
    title: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    is_initial: bool = True

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id = id or str(uuid.uuid4())
        return ScriptStage(
            id=_id,
            file=file,
            title=title,
            is_initial=True,
            workflow_transitions=[],
            workflow_position=workflow_position,
        )

    @staticmethod
    def from_dict(data: Dict):
        x, y = data["workflow_position"]
        return ScriptStage(
            id=data["id"],
            file=data["file"],
            title=data["title"],
            workflow_position=(x, y),
            is_initial=data["is_initial"],
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )

    @property
    def type_name(self):
        return "script"

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "type": "script",
            "is_initial": self.is_initial,
        }

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "file": self.file,
            "title": self.title,
            "is_initial": self.is_initial,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["id", "title", "schedule"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of script")

    def duplicate(self, new_id: str, new_position: Tuple[int, int]):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )


@dataclass
class JobStage:
    id: str
    file: str
    title: str
    schedule: str
    workflow_position: Tuple[float, float]
    workflow_transitions: List[WorkflowTransition]
    is_initial = True

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id = id or str(uuid.uuid4())
        return JobStage(
            id=_id,
            file=file,
            title=title,
            schedule="* * * * *",
            workflow_position=workflow_position,
            workflow_transitions=[],
        )

    @staticmethod
    def from_dict(data: Dict):
        x, y = data["workflow_position"]
        return JobStage(
            file=data["file"],
            id=data["id"],
            title=data["title"],
            schedule=data["schedule"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )

    @property
    def type_name(self):
        return "job"

    @property
    def admin_dto(self):
        return {"title": self.title, "id": self.id, "type": "job", "is_initial": True}

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "file": self.file,
            "title": self.title,
            "schedule": self.schedule,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    @property
    def editor_dto(self):
        return self.as_dict

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in ["identifier", "title", "schedule"]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            else:
                raise Exception(f"Cannot update {attr} of job")

    def duplicate(self, new_id: str, new_position: Tuple[int, int]):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )


@dataclass
class FormStage:
    id: str
    file: str
    title: str
    path: str
    workflow_transitions: List[WorkflowTransition]
    workflow_position: Tuple[float, float]
    notification_trigger: NotificationTrigger
    is_initial: bool = True
    end_message: Optional[str] = None
    auto_start: Optional[bool] = False
    start_message: Optional[str] = None
    error_message: Optional[str] = None
    welcome_title: Optional[str] = None
    allow_restart: Optional[bool] = False
    timeout_message: Optional[str] = None
    start_button_text: Optional[str] = None
    restart_button_text: Optional[str] = None

    @staticmethod
    def create(
        title: str,
        file: str,
        id: Union[str, None] = None,
        workflow_position: Tuple[int, int] = (0, 0),
    ):
        _id = id or str(uuid.uuid4())
        return FormStage(
            id=_id,
            file=file,
            path=to_kebab_case(file[:-3]),
            title=title,
            is_initial=True,
            workflow_transitions=[],
            workflow_position=workflow_position,
            notification_trigger=NotificationTrigger(
                variable_name="assignee_emails", enabled=False
            ),
        )

    @staticmethod
    def from_dict(data: dict):
        x, y = data["workflow_position"]
        return FormStage(
            id=data["id"],
            file=data["file"],
            path=data["path"],
            title=data["title"],
            end_message=data["end_message"],
            auto_start=data["auto_start"],
            start_message=data["start_message"],
            error_message=data["error_message"],
            welcome_title=data["welcome_title"],
            allow_restart=data["allow_restart"],
            timeout_message=data["timeout_message"],
            start_button_text=data["start_button_text"],
            restart_button_text=data["restart_button_text"],
            workflow_position=(x, y),
            is_initial=data["is_initial"],
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
            notification_trigger=NotificationTrigger(
                variable_name=data["notification_trigger"]["variable_name"],
                enabled=data["notification_trigger"]["enabled"],
            ),
        )

    @property
    def type_name(self):
        return "form"

    @property
    def admin_dto(self):
        return {
            "title": self.title,
            "id": self.id,
            "path": self.path,
            "type": "form",
            "is_initial": self.is_initial,
        }

    @property
    def browser_runner_dto(self):
        allow_restart = self.allow_restart if self.is_initial else False

        return {
            "id": self.id,
            "path": self.path,
            "title": self.title,
            "is_initial": self.is_initial,
            "auto_start": self.auto_start,
            "allow_restart": allow_restart,
            "end_message": self.end_message,
            "start_message": self.start_message,
            "error_message": self.error_message,
            "welcome_title": self.welcome_title,
            "timeout_message": self.timeout_message,
            "start_button_text": self.start_button_text,
            "restart_button_text": self.restart_button_text,
        }

    @property
    def editor_dto(self):
        return {
            **self.browser_runner_dto,
            "file": self.file,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
            "notification_trigger": (
                self.notification_trigger.to_dto()
                if self.notification_trigger
                else None
            ),
        }

    @property
    def as_dict(self):
        return self.editor_dto

    @property
    def file_path(self):
        return Settings.root_path.joinpath(self.file)

    def update(self, changes: Dict[str, Any]):
        for attr, value in changes.items():
            if attr in [
                "title",
                "end_message",
                "auto_start",
                "start_message",
                "error_message",
                "welcome_title",
                "allow_restart",
                "timeout_message",
                "start_button_text",
                "restart_button_text",
            ]:
                setattr(self, attr, value)
            elif attr == "file":
                _update_file(self, value)
            elif attr == "path":
                setattr(self, attr, normalize_path(value))
            elif attr == "notification_trigger":
                self.notification_trigger = NotificationTrigger(
                    variable_name=value["variable_name"], enabled=value["enabled"]
                )
            else:
                raise Exception(f"Cannot update {attr} of form")

    def duplicate(self, new_id: str, new_position: Tuple[int, int]):
        return self.from_dict(
            {
                **self.as_dict,
                "id": new_id,
                "workflow_position": new_position,
                "transitions": [],
            }
        )


@dataclass
class SidebarItem:
    id: str
    name: str
    path: str
    type: str
    visible: Optional[bool]

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "path": self.path,
            "type": self.type,
            "visible": self.visible,
        }


@dataclass
class Sidebar:
    items: List[SidebarItem]

    @staticmethod
    def from_dict(
        sidebar_data: List[dict],
        forms: List[FormStage] = [],
    ):
        def item_name(path):
            return [s.title for s in forms if s.path == path][0]

        stored_items = []

        initial_forms = [f for f in forms if f.is_initial]

        for item in sidebar_data:
            if item["path"] in [f.path for f in initial_forms]:
                sidebar_item = SidebarItem(
                    id=item["id"],
                    name=item_name(item["path"]),
                    path=item["path"],
                    type=item["type"],
                    visible=item.get("visible"),
                )
                stored_items.append(sidebar_item)

        for form in initial_forms:
            if form.path not in [item.path for item in stored_items]:
                stored_items.append(
                    SidebarItem(
                        id=form.id,
                        name=form.title,
                        path=form.path,
                        type="form",
                        visible=False,
                    )
                )
        return Sidebar(items=stored_items)

    @property
    def as_dict(self):
        return [item.as_dict for item in self.items]


@dataclass
class StyleSettings:
    name: str
    sidebar: Sidebar
    theme: Optional[str] = None
    logo_url: Optional[str] = None
    brand_name: Optional[str] = None
    main_color: Optional[str] = None
    font_family: Optional[str] = None
    font_color: Optional[str] = None

    @property
    def as_dict(self):
        return {
            "id": None,
            "name": self.name,
            "theme": self.theme,
            "logo_url": self.logo_url,
            "brand_name": self.brand_name,
            "main_color": self.main_color,
            "font_color": self.font_color,
            "font_family": self.font_family,
            "sidebar": self.sidebar.as_dict,
        }

    @property
    def browser_runner_dto(self):
        if isinstance(self.logo_url, str) and check_is_url(self.logo_url):
            logo_url = self.logo_url
        elif self.logo_url:
            logo_url = "/_assets/logo"
        else:
            logo_url = None
        return {**self.as_dict, "logo_url": logo_url}

    @property
    def editor_dto(self):
        return self.as_dict

    def update(
        self,
        changes: Dict[str, Any],
        forms: List[FormStage] = [],
    ):
        for attr, value in changes.items():
            if attr == "sidebar":
                self.sidebar = Sidebar.from_dict(changes["sidebar"], forms)
            else:
                setattr(self, attr, value)

    @staticmethod
    def from_dict(data: Dict, forms: List[FormStage] = []):
        return StyleSettings(
            sidebar=Sidebar.from_dict(data.get("sidebar", []), forms=forms),
            name=data.get("name", "Untitled Workspace"),
            font_family=data.get("font_family"),
            brand_name=data.get("brand_name"),
            main_color=data.get("main_color"),
            font_color=data.get("font_color"),
            logo_url=data.get("logo_url"),
            theme=data.get("theme"),
        )


class PathConflictError(Exception):
    pass


class StageNotFoundError(KeyError):
    pass


class TransitionNotFoundError(KeyError):
    pass


@dataclass
class VisualizationItem:
    name: str
    type: str

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, VisualizationItem):
            return False
        return self.name == __value.name and self.type == __value.type

    @staticmethod
    def from_dict(data: dict):
        return VisualizationItem(
            name=data["name"],
            type=data["type"],
        )

    @property
    def as_dict(self):
        return {
            "name": self.name,
            "type": self.type,
        }


@dataclass
class VisualizationSettings:
    items: List[VisualizationItem]

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, VisualizationSettings):
            return False
        return self.items == __value.items

    @staticmethod
    def create():
        return VisualizationSettings([])

    @staticmethod
    def from_dict(data: list):
        return VisualizationSettings(
            items=[VisualizationItem.from_dict(item) for item in data]
        )

    @property
    def as_dict(self):
        return [item.as_dict for item in self.items]


@dataclass
class IteratorStage:
    id: str
    title: str
    variable_name: str
    workflow_position: Tuple[float, float]
    workflow_transitions: List[WorkflowTransition]
    type_name: str = "iterator"

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "title": self.title,
            "variable_name": self.variable_name,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    def get_items(self, thread_data: Dict[str, Any]) -> List:
        variable_name = self.variable_name
        if not variable_name:
            return []

        raw_value = nested_get(thread_data, variable_name)
        if not isinstance(raw_value, list):
            return []

        return raw_value

    @staticmethod
    def from_dict(data: dict):
        x, y = data["workflow_position"]
        return IteratorStage(
            id=data["id"],
            title=data["title"],
            variable_name=data["variable_name"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )


@dataclass
class ConditionStage:
    id: str
    title: str
    variable_name: str
    workflow_position: Tuple[float, float]
    workflow_transitions: List[WorkflowTransition]
    type_name: str = "condition"

    @property
    def as_dict(self):
        return {
            "id": self.id,
            "title": self.title,
            "variable_name": self.variable_name,
            "workflow_position": self.workflow_position,
            "transitions": [t.as_dict for t in self.workflow_transitions],
        }

    def get_condition(self, thread_data: Dict[str, Any]) -> Any:
        variable_name = self.variable_name
        if not variable_name:
            return None

        return nested_get(thread_data, variable_name)

    @staticmethod
    def from_dict(data: dict):
        x, y = data["workflow_position"]
        return ConditionStage(
            id=data["id"],
            title=data["title"],
            variable_name=data["variable_name"],
            workflow_position=(x, y),
            workflow_transitions=[
                WorkflowTransition.from_dict(t) for t in data["transitions"]
            ],
        )


@dataclass
class Project:
    workspace: StyleSettings
    visualization: VisualizationSettings
    scripts: List[ScriptStage]
    forms: List[FormStage]
    hooks: List[HookStage]
    jobs: List[JobStage]
    iterators: List[IteratorStage]
    conditions: List[ConditionStage]

    _graph: Graph

    @property
    def as_dict(self):
        target_stages = set()
        for stage in self.workflow_stages:
            for transition in stage.workflow_transitions:
                target_stages.add(transition.target_id)

        for stage in self.jobs + self.forms + self.scripts + self.hooks:
            if stage.id in target_stages:
                stage.is_initial = False
            else:
                stage.is_initial = True

        return {
            "workspace": self.workspace.as_dict,
            "visualization": self.visualization.as_dict,
            "jobs": [job.as_dict for job in self.jobs],
            "hooks": [hook.as_dict for hook in self.hooks],
            "forms": [form.as_dict for form in self.forms],
            "scripts": [script.as_dict for script in self.scripts],
            "iterators": [i.as_dict for i in self.iterators],
            "conditions": [c.as_dict for c in self.conditions],
        }

    @property
    def actions(self) -> List[ActionStage]:
        return [*self.forms, *self.jobs, *self.hooks, *self.scripts]

    @property
    def workflow_stages(self) -> List[WorkflowStage]:
        return [*self.actions, *self.iterators, *self.conditions]

    def iter_entrypoints(self) -> Generator[Path, None, None]:
        for stage in self.actions:
            yield Path(stage.file)

    @property
    def project_files(self) -> Generator[Path, None, None]:
        for entrypoint in self.iter_entrypoints():
            yield from traverse_code(entrypoint)

    @property
    def project_local_dependencies(self) -> Generator[Path, None, None]:
        entrypoints = list(self.iter_entrypoints())
        for file in self.project_files:
            if file not in entrypoints:
                yield file

    def get_next_stages_ids(self, id: str) -> List[str]:
        return [node.id for node in self._graph.next_to(id)]

    def get_previous_stages_ids(self, id: str) -> List[str]:
        return [node.id for node in self._graph.previous_to(id)]

    def is_initial(self, target_stage: ActionStage) -> bool:
        return len(self._graph.previous_to(target_stage.id)) == 0

    def add_stage(self, stage: WorkflowStage):
        if isinstance(stage, FormStage):
            self.forms.append(stage)
        elif isinstance(stage, HookStage):
            self.hooks.append(stage)
        elif isinstance(stage, JobStage):
            self.jobs.append(stage)
        elif isinstance(stage, ScriptStage):
            self.scripts.append(stage)
        elif isinstance(stage, IteratorStage):
            self.iterators.append(stage)
        elif isinstance(stage, ConditionStage):
            self.conditions.append(stage)
        else:
            raise Exception(f"Cannot add stage of type {type(stage)}")

    def get_stage(self, id: str) -> Optional[WorkflowStage]:
        for stage in self.workflow_stages:
            if stage.id == id:
                return stage

        return None

    def get_stage_raises(self, id: str) -> WorkflowStage:
        stage = self.get_stage(id)
        if not stage:
            raise StageNotFoundError(f"Stage with id '{id}' not found")
        return stage

    def get_action(self, id: str) -> Optional[ActionStage]:
        for stage in self.actions:
            if stage.id == id:
                return stage

        return None

    def get_action_raises(self, id: str) -> ActionStage:
        stage = self.get_action(id)
        if not stage:
            raise StageNotFoundError(f"Stage with id '{id}' not found")
        return stage

    def get_form(self, id: str) -> Optional[FormStage]:
        for form in self.forms:
            if form.id == id:
                return form

        return None

    def get_hook(self, id: str) -> Optional[HookStage]:
        for hook in self.hooks:
            if hook.id == id:
                return hook

        return None

    def get_job(self, id: str) -> Optional[JobStage]:
        for job in self.jobs:
            if job.id == id:
                return job

        return None

    def get_script(self, id: str) -> Optional[ScriptStage]:
        for script in self.scripts:
            if script.id == id:
                return script

        return None

    def get_hook_by_path(self, path: str) -> Optional[HookStage]:
        for hook in self.hooks:
            if hook.path == path:
                return hook

        return None

    def get_form_by_path(self, path: str) -> Optional[FormStage]:
        for form in self.forms:
            if form.path == path:
                return form

    def get_stage_by_path(self, path: str) -> Optional[ActionStage]:
        for form in self.forms:
            if form.path == path:
                return form

        for hook in self.hooks:
            if hook.path == path:
                return hook

        return None

    def get_initial_stages(self) -> List[ActionStage]:
        return [r for r in self.actions if self.is_initial(r)]

    def delete_transition_by_target(self, target_id: str):
        for stage in self.workflow_stages:
            stage.workflow_transitions = [
                t for t in stage.workflow_transitions if t.target_id != target_id
            ]

    def update_path(self, stage: ServedStage, new_path: str):
        old_path = stage.path

        if self.get_stage_by_path(new_path):
            raise PathConflictError(f"Path {new_path} already exists")

        all_stages: List[Union[FormStage, HookStage, JobStage]] = [
            *self.forms,
            *self.hooks,
            *self.jobs,
        ]

        for _ in all_stages:
            for transition in stage.workflow_transitions:
                if transition.target_id == old_path:
                    transition.target_id = new_path

        stage.path = new_path

    def update_stage(self, stage: ActionStage, changes: Dict[str, Any]) -> ActionStage:
        new_path = changes.get("path")

        if new_path:
            if not isinstance(stage, FormStage) and not isinstance(stage, HookStage):
                raise Exception(
                    f"Stage with id {id} is a {type(stage)} does not have a path"
                )

            normalized_path = normalize_path(new_path)
            self.update_path(stage, normalized_path)

        stage.update(changes)

        return stage

    def delete_stage(self, id: str, remove_file: bool = False):
        if remove_file:
            stage = self.get_action(id)
            if not stage:
                return
            path = Settings.root_path.joinpath(stage.file)
            if path.exists():
                os.remove(path.absolute())

        self.delete_transition_by_target(id)
        self.forms = [f for f in self.forms if f.id != id]
        self.hooks = [h for h in self.hooks if h.id != id]
        self.jobs = [j for j in self.jobs if j.id != id]
        self.scripts = [s for s in self.scripts if s.id != id]
        self.iterators = [i for i in self.iterators if i.id != id]
        self.conditions = [c for c in self.conditions if c.id != id]

    @staticmethod
    def from_dict(data: dict):
        target_stages = set()
        nodes = []
        edges = []

        control_keys = ["conditions", "iterators"]
        stage_keys = ["forms", "hooks", "scripts", "jobs"]

        for key in stage_keys + control_keys:
            for stage in data[key]:
                nodes.append(Node(id=stage["id"]))
                for transition in stage.get("transitions", []):
                    target_stages.add(transition.get("target_id"))
                    edges.append(
                        Edge(source_id=stage["id"], target_id=transition["target_id"])
                    )

        for key in stage_keys:
            for index, stage in enumerate(data[key]):
                if stage["id"] in target_stages:
                    data[key][index]["is_initial"] = False
                else:
                    data[key][index]["is_initial"] = True

        try:
            scripts = [ScriptStage.from_dict(script) for script in data["scripts"]]
            forms = [FormStage.from_dict(form) for form in data["forms"]]
            hooks = [HookStage.from_dict(hook) for hook in data["hooks"]]
            jobs = [JobStage.from_dict(job) for job in data["jobs"]]
            iterators = [IteratorStage.from_dict(i) for i in data["iterators"]]
            conditions = [ConditionStage.from_dict(c) for c in data["conditions"]]
            visualization = VisualizationSettings.from_dict(data["visualization"])

            workspace = StyleSettings.from_dict(data["workspace"], forms=forms)

            return Project(
                workspace=workspace,
                visualization=visualization,
                scripts=scripts,
                forms=forms,
                hooks=hooks,
                jobs=jobs,
                iterators=iterators,
                conditions=conditions,
                _graph=Graph.from_primitives(nodes=nodes, edges=edges),
            )

        except Exception:
            print("Error: incompatible abstra.json file.")
            import traceback

            traceback.print_exc()
            sys.exit(1)

    @staticmethod
    def create():
        return Project(
            workspace=StyleSettings(name="Untitled Workspace", sidebar=Sidebar([])),
            visualization=VisualizationSettings.create(),
            scripts=[],
            forms=[],
            hooks=[],
            jobs=[],
            iterators=[],
            conditions=[],
            _graph=Graph.from_primitives([], []),
        )


def _update_file(
    stage: ActionStage,
    new_file_relative: str,
):
    old_file = Settings.root_path.joinpath(stage.file)
    new_file = Settings.root_path.joinpath(new_file_relative)

    if old_file.exists() and not new_file.exists():
        new_file.parent.mkdir(parents=True, exist_ok=True)
        old_file.rename(new_file)

    stage.file = new_file_relative


class ProjectRepository:
    @classmethod
    def get_file_path(cls):
        return Settings.root_path / "abstra.json"

    @classmethod
    def initialize(cls):
        if not cls.exists():  # double check
            cls.save(Project.create())

    @classmethod
    def exists(cls):
        return cls.get_file_path().exists()

    @classmethod
    def save(cls, project: Project):
        temp_file = Path(tempfile.mkdtemp()) / "abstra.json"

        project_data = project.as_dict

        project_data["version"] = json_migrations.get_latest_version()

        with temp_file.open("w") as f:
            json.dump(project_data, f, indent=2)

        shutil.move(str(temp_file), cls.get_file_path())
        Path.rmdir(temp_file.parent)

    @classmethod
    def migrate_config_file(cls):
        if not cls.exists():
            return
        data = json.loads(cls.get_file_path().read_text(encoding="utf-8"))
        initial_version = data.get("version")

        migrated_data = json_migrations.migrate(
            data,
            Settings.root_path,
        )

        if migrated_data["version"] != initial_version:
            cls.save(Project.from_dict(migrated_data))

    @classmethod
    def load(cls) -> Project:
        data = json.loads(cls.get_file_path().read_text(encoding="utf-8"))

        return Project.from_dict(data)

    @classmethod
    def initialize_or_migrate(cls):
        if not cls.exists():
            cls.initialize()
        else:
            cls.migrate_config_file()



import re
from unidecode import unidecode


def to_kebab_case(s: str):
    lower = s.lower().strip()
    unaccented = unidecode(lower)
    alphanumeric_pattern = re.compile(r"[a-z0-9]+")
    matches = alphanumeric_pattern.findall(unaccented)
    return "-".join(matches) if matches else ""



import sys
from typing import Callable, Literal, Union

from .env_masker import GLOBAL_MASKER
from .execution.execution import Execution
from .stdio_skip_request import ESCAPE_STRING
from .utils.environment import IS_PRODUCTION

_og_stdout_write = sys.stdout.write
_og_stderr_write = sys.stderr.write


def writeWraper(
    type: Literal["stdout", "stderr"],
    write: Callable,
    raw: Union[str, bytearray],
    print_exceptions: bool,
):
    text = raw.decode("utf-8") if not isinstance(raw, str) else raw

    if IS_PRODUCTION:
        text = GLOBAL_MASKER.mask(text)
    try:
        skip_sending = text.startswith(ESCAPE_STRING)
        execution = Execution.get_current_execution()

        term_output = text

        if execution and not skip_sending:
            term_output = (
                f"[RUN {execution.id.split(sep='-')[0]}] {term_output}"
                if term_output.strip()
                else term_output
            )
            execution.stdio(type, text)
        write(term_output)
    except Exception as e:
        if print_exceptions:
            write(f"[STDIO_WRAPPER_ERROR]: {e}")
    finally:
        sys.stdout.flush()
        return len(text)


def override_stdio(print_exceptions: bool):
    sys.stdout.write = lambda raw: writeWraper(
        "stdout", _og_stdout_write, raw, print_exceptions
    )
    sys.stderr.write = lambda raw: writeWraper(
        "stderr", _og_stderr_write, raw, print_exceptions
    )


def reset():
    sys.stdout.write = _og_stdout_write
    sys.stderr.write = _og_stderr_write



from unittest import TestCase
from .comparators import (
    is_eq,
    is_neq,
    is_gt,
    is_between,
    is_gte,
    is_in,
    is_lt,
    is_like,
    is_lte,
    is_not_in,
    is_not_like,
    is_null,
    is_not_null,
    is_ilike,
    is_not_ilike,
)


class TestComparators(TestCase):
    def test_is_eq(self):
        self.assertEqual(is_eq(1).exp("column", 1), ('"column"=$1', [1]))

    def test_is_neq(self):
        self.assertEqual(is_neq(1).exp("column", 1), ('"column" <> $1', [1]))

    def test_is_gt(self):
        self.assertEqual(is_gt(1).exp("column", 1), ('"column" > $1', [1]))

    def test_is_between(self):
        self.assertEqual(
            is_between(1, 2).exp("column", 1), ('"column" BETWEEN $1 AND $2', [1, 2])
        )

    def test_is_gte(self):
        self.assertEqual(is_gte(1).exp("column", 1), ('"column" >= $1', [1]))

    def test_is_in(self):
        self.assertEqual(is_in([1, 2]).exp("column", 1), ('"column" IN $1', [[1, 2]]))

    def test_is_lt(self):
        self.assertEqual(is_lt(1).exp("column", 1), ('"column" < $1', [1]))

    def test_is_like(self):
        self.assertEqual(is_like("a").exp("column", 1), ('"column" LIKE $1', ["a"]))

    def test_is_lte(self):
        self.assertEqual(is_lte(1).exp("column", 1), ('"column" <= $1', [1]))

    def test_is_not_in(self):
        self.assertEqual(
            is_not_in([1, 2]).exp("column", 1), ('"column" NOT IN $1', [[1, 2]])
        )

    def test_is_not_like(self):
        self.assertEqual(
            is_not_like("a").exp("column", 1), ('"column" NOT LIKE $1', ["a"])
        )

    def test_is_null(self):
        self.assertEqual(is_null().exp("column", 1), ('"column" IS NULL', []))

    def test_is_not_null(self):
        self.assertEqual(is_not_null().exp("column", 1), ('"column" IS NOT NULL', []))



from .base_migration import Migration


class Migration004(Migration):
    @staticmethod
    def target_version() -> str:
        return "4.0"

    @staticmethod
    def source_version() -> str:
        return "3.0"

    def _migrate(self) -> None:
        forms = self.data.get("forms", [])

        for form in forms:
            form["notification_trigger"] = {
                "variable_name": "assignee_emails",
                "enabled": False,
            }
